library("RMVL")
library("lattice")
library("parallel")

cat("Starting parallel cluster\n")
# Use 4 threads
setDefaultCluster(makeForkCluster(4, useXDR=FALSE))

cat("Using", length(getDefaultCluster()), "parallel R instances\n")

cat("Loading data\n")

Matlas<-mvl_open("O3a_atlas.mvl")
Msummary<-mvl_open("O3a_atlas_summary.mvl")

# Open datasets on for each parallel R instance
# Memory mapped data is shared

clusterEvalQ(cl=NULL, {
	library("RMVL")
	Matlas<-mvl_open("O3a_atlas.mvl")
	Msummary<-mvl_open("O3a_atlas_summary.mvl")
	})

outliers<-read.table("O3_outliers.csv", header=TRUE)
outliers[,"RAJD"]<-outliers[,"ra"]*180/pi
outliers[,"DECJD"]<-outliers[,"dec"]*180/pi

visualize_help<-function() {

	cat("\n\tskymap(fstart, fstop, metric)\n")
	cat("\tplot_data(fstart, fstop, metric, mark.rad, mark.deg, mark.col)\n")
	cat("\n")
	cat("The function skymap() returns a data.frame with sky map data, while plot_data() plots it.\n")
	cat("Parameters:\n")
	cat("\tfstart      - frequency band start\n")
	cat("\tfstop       - frequency band stop\n")
	cat("\tmetric      - either \"snr_max\" (for SNR) or \"ul_max\" (for worst-case upper limit)\n")
	cat("\tmark.rad    - coordinates of optional mark to add to the plot c(ra=RA, dec=Dec), in radians, J2000\n")
	cat("\tmark.deg    - coordinates of optional mark to add to the plot c(ra=RA, dec=Dec), in degrees, J2000\n")
	cat("\tmark.col    - color of optional mark, defaults to \"red\"\n")
	
	cat("\nExamples:\n\n")
	cat("\tplot_data(835, 840, \"snr_max\")\n\n")
	cat("\tplot_data(835, 840, \"ul_max\", mark.deg=c(ra=45, dec=0), mark.col=\"magenta\")\n\n")
	cat("\tdf<-skymap(835, 840, \"snr_max\")\n\n")
	}

Lines<-Msummary$lines[,]

epsilon<-23.439281*pi/180
# x, y, z
ecliptic_pole<-c(0, -sin(epsilon), cos(epsilon))

galactic_ra<-(12+51.4/60)*(2*pi)/24
galactic_dec<-27.17*pi/180
galactic_pole<-c(cos(galactic_dec)*cos(galactic_ra), cos(galactic_dec)*sin(galactic_ra), sin(galactic_dec))

plot_x<-function(ra, dec) { return(180+(180-ra)*cos(dec*pi/180.0)) }
limit_y<-function(y) { return(ifelse(abs(y)>90, NA, y)) } 

plot_band<-function(v, width, ...) {
	v<-v/sqrt(sum(v*v))
	ra0<-seq(0, 360, by=1)

	x1<-cos(ra0*pi/180)*v[1]
	y1<-sin(ra0*pi/180)*v[2]
	a1<-x1+y1
	z1<-v[3]
	
	w1a<- cos(pi/2-(0.5*width/180)*pi)
	w1b<- cos(pi/2+(0.5*width/180)*pi)
	
	n1<-sqrt(a1^2+z1^2)
	
	a2<-a1/n1
	z2<-z1/n1
	w2a<-w1a/n1
	w2b<-w1b/n1
	
	angle3<-atan2(z2, a2)*180/pi
	dec0a<- acos(w2a)*180/pi+angle3
	dec0b<- acos(w2b)*180/pi+angle3
	
	points(plot_x(ra0, dec0a)/360, 0.5+limit_y(dec0a)/180, type="l", ...)
	points(plot_x(ra0, dec0b)/360, 0.5+limit_y(dec0b)/180, type="l", ...)
	
	dec0a<- -acos(w2a)*180/pi+angle3
	dec0b<- -acos(w2b)*180/pi+angle3
	
	points(plot_x(ra0, dec0a)/360, 0.5+limit_y(dec0a)/180, type="l", ...)
	points(plot_x(ra0, dec0b)/360, 0.5+limit_y(dec0b)/180, type="l", ...)
	}

plot_points<-function(RAJD, DECJD, ...) {
	points(plot_x(RAJD, DECJD)/360, 0.5+limit_y(DECJD)/180, ...)
	}

plot_ecliptic_band<-function(width, ...) {
	plot_band(ecliptic_pole, width, ...)
	}

plot_galactic_band<-function(width, ...) {
	plot_band(galactic_pole, width, ...)
	}

atlas_max<-function(fstart, fstop, col, idx=NULL) {
	params<-Matlas$parameters
	if(is.null(idx)) {
		idx<-params[params[,"band_start"][]+0.045>=fstart & (params[,"band_start"][]<fstop), "idx"]
		}

	cv<-Matlas$skymaps[,col,ref=TRUE]
		
	idx2<-unlist(mvl_index_lapply(Matlas["skymaps_extent_index_idx"], list(idx), function(i, idx) {
		v<-cv[idx]
		return(idx[v==max(v)])
		}))
	idx3<-max(Matlas$skymaps[idx2, col])==Matlas$skymaps[idx2, col]
	sk<-Matlas$skymaps[idx2[idx3],,drop=FALSE]
	sk[,"RAJD"]<-sk[,"ra"]*180/pi
	sk[,"DECD"]<-sk[,"dec"]*180/pi
	return(sk)
	}

band_plan<-data.frame(band_start=sort(unique(Matlas$parameters[,"band_start"][])))
band_plan[,"idx"]<-1:(dim(band_plan)[1])
band_plan[,"idx2"]<-((band_plan[,"idx"]-1) %/% 32)+1
band_plan[,"idx3"]<-((band_plan[,"idx"]-1) %/% 1024)+1
	
full_spectrum<-mvl2R(Msummary$spectrum)

clusterExport(cl=NULL, "full_spectrum")

get_grid<-function() {
	G<-Msummary$summary_group_sky
	first_idx<-G$index[G$stretch_index[1:(length(G$stretch_index)-1)]]
	return(Msummary$summary[first_idx, c("ra", "dec", "x", "y")])
	}

full_grid<-get_grid()

split_groups<-function(G, N) {
	L<-lapply(seq(1, length(G$stretch_index)-1, by=N), function(i) {
		idx<-i+(0:N)
		idx<-idx[idx<=length(G$stretch_index)]
		return(G$stretch_index[idx])
		})
	return(L)
	}

skymap<-function(fstart=0, fend=1e5, col) {

	M0<-clusterApply(cl=NULL, split_groups(Msummary$summary_group_sky, 100), function(si) {
			G<-list(stretch_index=si-(si[1]-1), index=Msummary$summary_group_sky$index[si[1]:si[length(si)]])
	
			data<-Msummary$summary

			d<-data[, col]
			
			band_start<-data[, "band_start"]

			idx_max<-function(idx) {
					bs<-band_start[idx]
					Fv<-(bs+0.045>=fstart) & (bs<=fend)
					idx2<-idx[Fv]
					return(max(d[idx2]))
					}
			
			return(unlist(mvl_group_lapply(G, idx_max)))
			})
	
	M<-do.call(c, M0)
	
	df<-data.frame(full_grid, M)
	
	names(df)[dim(df)[2]]<-col
			
	return(df[order(df[,"dec"], df[,"ra"]), ])
	}
	
# formatting function
# https://stackoverflow.com/questions/10762287/how-can-i-format-axis-labels-with-exponents-with-ggplot2-and-scales
scientific <- function(x) {
    if(any(regexpr("e", x)>=0))return(ifelse(x==0, "0", parse(text=gsub("[+]", "", gsub("e", " %*% 10^", x)))))
	else return(x)
}

color.bar<-function(lut, min, max=-min, nticks=11, ticks=NULL, title=NA, ...) {
	scale <- (length(lut)-1)/(max-min)
	
	if(is.null(ticks)) {
		ticks<-pretty(c(min, max), n=nticks)
		ticks<-c(min, ticks[ticks>min & ticks<max], max)
		
		ticks<-seq(min, max, length=nticks)
		}

	plot(c(0,10), c(min,max), type='n', bty='n', xaxt='n', xlab=NA, yaxt='n', ylab=NA, main=title)
	axis(2, at=ticks, labels=scientific(format(ticks)), las=1, ...)
	for(i in 1:(length(lut)-1)) {
		y<-(i-1)/scale + min
		rect(0,y,10,y+1/scale, col=lut[i], border=NA)
		}
	}

custom_legend<-function(title=NA, ...) {
	plot(c(0,100), c(0,100), type='n', bty='n', xaxt='n', xlab=NA, yaxt='n', ylab=NA, main=title)
	legend(0, 100, c("outlier", "line", "ecl_max"), col=c("blue", "red", "magenta"), pch=c(5, 8, 1))
	}
		
plot_skymap<-function(sk, col, levels=NULL, ...) {
	N_x<-max(sk[,"x"])
	N_y<-max(sk[,"y"])
	M<-matrix(NA, N_x, N_y)
 	for(i in 1:(dim(sk)[1]))
 		M[sk[i,"x"], sk[i,"y"]]<-sk[i,col]
 		
	cols<-rev(rainbow(50, end=0.75))
	image(M, col=cols, xaxt= "n", yaxt= "n", xlab="RA", ylab="Dec", ...)
	
	axis(1, at=seq(0, 1, 1/12), labels=rev(seq(0, 360, by=30)), ...)
	axis(2, at=seq(0, 1, 1/6), labels=(seq(-90, 90, by=30)), ...)
	
	}
	
rational_levels<-function(x) {
	levels<-quantile(x, c(0, 1))
	M<-max(abs(levels-mean(levels)))
	if(M>0) {
		M1<-10^(ceiling(log10(2*M))-2)
		levels[1]<-floor(levels[1]/M1)*M1
		levels[2]<-ceiling(levels[2]/M1)*M1
		}
	return(levels)
	}

	
plot_data<-function(fstart=0, fend=1e5, col, galactic.band=TRUE, plot.outliers=TRUE, mark.rad=NULL, mark.deg=NULL, mark.col="red", ...) {
	if(! col %in% c("snr_max", "ul_max")) stop("Parameter col=\"", col, "\" must be either \"snr_max\" or \"ul_max\"")

	sk<-skymap(fstart=fstart, fend=fend, col=col)
	layout(matrix(c(1,2,3,4), 2, 2, byrow = TRUE), width=c(17, 3), heights=c(2,1))
	plot_skymap(sk, col, main=ifelse(col=="snr_max", expression("SNR"), expression(paste(h[0], " upper limit"))), ...)
	
	if(galactic.band)
		plot_galactic_band(25, col="red", lwd=2)
		
	if(plot.outliers) {
		Fp<-outliers[,"snr"]>0
		Fp[is.na(Fp)]<-TRUE
		Fp<-Fp & outliers[,"frequency"]>=fstart & outliers[,"frequency"]<=fend+0.045
		Fp[is.na(Fp)]<-FALSE
		if(any(Fp)) {
			cat(sum(Fp), "outliers in band\n")
			Pdf<-outliers[Fp,,drop=FALSE]
			if(dim(Pdf)[1]<10)print(Pdf)
			plot_points(Pdf[,"RAJD"], Pdf[,"DECJD"], pch=5, col="blue", cex=2)
			}
		}
		
	if(!is.null(mark.deg)) {
		plot_points(mark.deg[["ra"]], mark.deg[["dec"]], pch=5, col=mark.col, cex=2)
		}
		
	if(!is.null(mark.rad)) {
		plot_points(mark.rad[["ra"]]*180/pi, mark.rad[["dec"]]*180/pi, pch=5, col=mark.col, cex=2)
		}
	
	pp<-par(mar=c(2, 4, 2, 1))
	
	
	levels<-rational_levels(sk[,col])
 		
	color.bar(rev(rainbow(50, end=0.75)), levels[[1]], levels[[2]], ...)
	par(pp)
	
	pp<-par(mar=c(5, 4, 1, 2)+0.1)
	sp<-full_spectrum
	Fband<-sp[,"band_start"]>=fstart & sp[,"band_start"]<=fend
	bnd<-rational_levels(sp[Fband, col])
	if(sum(Fband)>20) {
		plot(sp[Fband,"band_start"], sp[Fband, col], type="l", xlab="Frequency, Hz", ylab=col, yaxt="n", col="black", ylim=bnd,  ...)
		X<-sp[Fband, col]
		X[sp[Fband, "line_mark"]<1]<- NA
		points(sp[Fband,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, pch=8, col="red", ...)
		
		X<-sp[Fband, col]
		X[sp[Fband, "snr_max"]>sp[Fband, "snr_ecl"]]<- NA
		points(sp[Fband,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, col="magenta", ...)
		
		}
		else {
		plot(sp[Fband,"band_start"], sp[Fband, col], type="o", xlab="Frequency, Hz", ylab=col, yaxt="n", col="black", ylim=bnd,  ...)

		X<-sp[Fband, col]
		X[sp[Fband, "line_mark"]<1]<- NA
		points(sp[Fband,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, pch=8, col="red", ...)
		
		X<-sp[Fband, col]
		X[sp[Fband, "snr_max"]>sp[Fband, "snr_ecl"]]<- NA
		points(sp[Fband,"band_start"], X, type="o", xlab="Frequency, Hz", ylab=col, col="magenta", ...)
		}
		
	ticks<-seq(bnd[[1]], bnd[[2]], length=5)
	axis(2, at=ticks, labels=ifelse(c(TRUE, rep(FALSE, length(ticks)-2), TRUE),scientific(format(ticks)), NA))
		
	if(plot.outliers) {
		X<-sp[Fband, col]
		FF<-sp[Fband, "band_start"]
		Fn0<-outliers[,"snr"]>0
		Fn0[is.na(Fn0)]<-TRUE
		if(length(X)>0)
		for(i in 1:(length(X))) {
			Fn<-Fn0 & FF[i]<= outliers[,"frequency"] & FF[i]+0.045>=outliers[,"frequency"]
			Fn[is.na(Fn)]<-FALSE
			if(!any(Fn))X[i]<-NA
			}
		points(sp[Fband,"band_start"], X, type="o", pch=5, col="blue", ...)
		}
	par(pp)
		
	pp<-par(mar=c(0, 0, 1, 1), mgp=c(0, 0, 0))
	custom_legend()
	par(pp)
		
	cat("Maximum of" ,col, "is achieved at:\n")
	idx<-which.max(sk[,col])
	sk2<-sk[idx,,drop=FALSE]
	sk2[,"RAJD"]<-180*sk2[,"ra"]/pi
	sk2[,"DECD"]<-180*sk2[,"dec"]/pi
	print(sk2)
	}
	
#
# Uncomment and adjust to match DPI setting of your screen
#
#Screen_DPI<-192	
#X11("", 10*Screen_DPI/96, 8*Screen_DPI/96, 12*Screen_DPI/96)

cat("Producing sample plot\n")

plot_data(835, 840, "snr_max")


# Change to TRUE to produce skymaps from the paper
if(FALSE) {

# png("skymap_835_840_snr.png", width=1280, height=1024, res=125)
# plot_data(835, 840, "snr_max")
# dev.off()
# 
# png("skymap_835_840_ul.png", width=1280, height=1024, res=125)
# plot_data(835, 840, "ul_max")
# dev.off()

pdf("skymap_835_840_snr.pdf", width=12, height=10, pointsize=20)
plot_data(835, 840, "snr_max")
dev.off()

pdf("skymap_835_840_ul.pdf", width=12, height=10, pointsize=20)
plot_data(835, 840, "ul_max")
dev.off()

# pdf("skymap_575_575_25_ip2_ul.pdf", width=12, height=10)
# plot_data(575, 575.25, "ul_max")
# dev.off()

pdf("skymap_763_764_ip9_ul.pdf", width=12, height=10, pointsize=20)
plot_data(763, 764, "ul_max")
dev.off()

pdf("skymap_763_3_763_42_ul.pdf", width=12, height=10, pointsize=20)
plot_data(763.3, 763.42, "ul_max")
dev.off()
}

cat("\nCall visualize_help() for examples and usage\n\n")

